-----------------------------------------------------------------------------------------------
-- Client Lua Script for MBM
-- Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------
 
require "Window"
 
-----------------------------------------------------------------------------------------------
-- MBM Module Definition
-----------------------------------------------------------------------------------------------
local MBM = {}
local MRF = nil
 
-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------
-- e.g. local kiExampleVariableMax = 999
 
-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function MBM:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 

    -- initialize variables here

    return o
end

function MBM:Init()
	local bHasConfigureFunction = false
	local strConfigureButtonText = "MischhBuddyModule"
	local tDependencies = {"MischhRaidFrames"}
    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end
 

-----------------------------------------------------------------------------------------------
-- MBM OnLoad
-----------------------------------------------------------------------------------------------
function MBM:OnLoad()
    -- load our form file
	MBM = self
	self.xmlDoc = XmlDoc.CreateFromFile("MBM_xml.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
	
	--scince we set the dependency, we can assume MRF is loaded (aka has :OnLoad executed already)
	MRF = Apollo.GetAddon("MischhRaidFrames")
end

function MBM:LoadForm(name, parent, handler)
	return Apollo.LoadForm(self.xmlDoc, name, parent, handler)
end

-----------------------------------------------------------------------------------------------
-- MBM OnDocLoaded
-----------------------------------------------------------------------------------------------
do
	local load = {}
	local loaded = false
	
	local function exec(f, t)
		if not t then
			f()
		elseif type(t[f]) == "function" then
			t[f](t)
		end
	end
	
	function MBM:OnDocLoaded()
		loaded = true
		
		for f,t in pairs(load) do
			exec(f,t)
		end 
	end
	
	function MBM:OnceDocLoaded(func, tbl)
		if not func then return end
			
		if not loaded then
			if type(tbl) == "table" then
				load[func] = tbl --if tbl is actually a table, we assume func is a correct key for it.
			elseif type(func) == "function" then
				load[func] = false
			end --all other situations are useless.
		else
			if type(tbl) == "table" then
				if type(tbl[func]) == "function" then
					tbl[func](tbl)
				end
			elseif type(func) == "function" then
				func()
			end
		end
	end
end
-----------------------------------------------------------------------------------------------
-- MBM Core Functionality
-----------------------------------------------------------------------------------------------

MBM:OnceDocLoaded(function() MRF:OnceDocLoaded("OnLoadedDocs", MBM) end)
function MBM:OnLoadedDocs()
	local frame = MBM:LoadForm("Window", "InWorldHudStratum")
	local ModOpt = MRF:GetOption(nil, "BuddyModule")
	
	local units = MRF:GetUnits()
	local optActive = MRF:GetOption(ModOpt, "activated")
	
	local executePriorities = {
		"UpdatePlayerPos",				--needed for distance filter (if activated)
		"Undraw",						--Remove everything.
		"DrawNormal", "DrawNoNormal",	--first: draw everything non-special.
		"DrawNoFoci", "DrawFoci", 		--then draw the foci ontop
		"DrawPlayer",					--as the very last: draw the Players circle!
		"DrawSubdue",}					--Special stuff later..
	local executeOrder = {}
	local executeThis = {
		Undraw = true,
		DrawNormal = true,
		DrawFoci = true,
		DrawPlayer = true,
	}
	function MBM:Reorder()
		executeOrder = {}
		for _, key in ipairs(executePriorities) do
			if executeThis[key] then
				executeOrder[#executeOrder+1] = key
			end
		end
	end
	function MBM:OnFrame()
		for _, key in ipairs(executeOrder) do
			self[key](self)
		end
	end
	
	optActive:OnUpdate(function(act)
		if act == nil then
			optActive:Set(true)
		elseif act then
			Apollo.RegisterEventHandler("NextFrame", "OnFrame", MBM)
		else
			Apollo.RemoveEventHandler("NextFrame", MBM)
			MBM:Undraw()
		end
	end)
	
	do
		local checks = {}
		--this gets called whenever a color-option seems to be invalid at first peak
		--we wait a frame and then work out, if its actually the case.
		function MBM:CheckThisColorOpt(opt)
			if type(opt:Get()) ~= "table" then
				return opt:Set(MRF:GetDefaultColor())
			end		
			checks[#checks+1] = opt
			checks.timer = checks.timer or ApolloTimer.Create(0, false, "OnFrame_ColorCheck", self)
		end
		
		function MBM:OnFrame_ColorCheck()		
			local tmp = checks --invoking the colors could call :CheckThisColorOpt() - we do not want to use this table then.
			checks.timer:Stop()
			checks = {}
			for i, opt in ipairs(tmp) do
				local col = opt:Get()
				if type(col) ~= "table" or type(col.Get) ~= "function" then
					col = MRF:GetDefaultColor()
				end
				opt:Set(col) --we want to update the option in every case!
			end
		end
		
		local colors = {}
		function MBM:RegisterColorOption(opt)
			local lastCol = nil 
			local f = function(newCol)
				if lastCol then
					colors[lastCol] = nil
				end
				if newCol then
					colors[newCol] = opt
				end
				lastCol = newCol
			end
			f(opt:Get())
			
			opt:OnUpdate(f)
		end
		
		MRF:GetOption("Event_RemovedColor"):OnUpdate(function(col) 
			if colors[col] then
				colors[col]:Set(MRF:GetDefaultColor())
			end
		end)
	end
	
	function MBM:GetCutoff(vec2)
		return Vector2.New(0,0)
	end
	
	do
		local halfWidth, halfHeight = math.floor(frame:GetWidth()/2), math.floor(frame:GetHeight()/2)
		function MBM:UnitToVecs(unit)
			local pos = unit:GetPosition()
			if not pos or pos.x==0 then return end
			
			local scrVec = GameLib.WorldLocToScreenPoint(Vector3.New(pos))
			local tarVec = Vector2.New(scrVec.x-halfWidth, scrVec.y-halfHeight)
			
			return self:GetCutoff(tarVec), tarVec
		end
	end
	
	function MBM:Draw(startVec, endVec, col, width)
		frame:AddPixie({
			bLine = true, fWidth = width,  cr = col,
			loc = { fPoints = { 0.5, 0.5, 0.5, 0.5 }, nOffsets = { startVec.x, startVec.y, endVec.x, endVec.y}}
		})
	end
	
	function MBM:Undraw()
		frame:DestroyAllPixies()
	end

	do --subdue
		local L = MRF:Localize({
			["SubduePattern"] = "^(%S+)'s Weapon",
		},{
			["SubduePattern"] = "Waffe von (.+)$",
		},{
			["SubduePattern"] = "Arme de (.+)$",
		})
		local subUnit = nil
	
		local optSubdue = MRF:GetOption(ModOpt, "subdue")
		local optSubAct = MRF:GetOption(optSubdue, "activate")
		local color, optSubCol = nil, MRF:GetOption(optSubdue, "color"); MBM:RegisterColorOption(optSubCol)
		local width, optSubWid = nil, MRF:GetOption(optSubdue, "width")
		
		local function registerStuff(a,b)
			if a and b then
				Apollo.RegisterEventHandler("UnitCreated", "OnUnitCreated", MBM)
				Apollo.RegisterEventHandler("UnitDestroyed", "OnUnitDestroyed", MBM)
			else
				Apollo.RemoveEventHandler("UnitCreated", MBM)
				Apollo.RemoveEventHandler("UnitDestroyed", MBM)
			end
		end
		
		optActive:OnUpdate(function(act) --Caution! this is the whole Modules active-state!
			registerStuff(optSubAct:Get(),act)
		end)
		
		optSubAct:OnUpdate(function(act)
			if act == nil then
				optSubAct:Set(true)
			elseif act then
				registerStuff(act,optActive:Get())
				executeThis.DrawSubdue = true
				MBM:Reorder()
			else
				registerStuff(act,optActive:Get())
				executeThis.DrawSubdue = nil
				MBM:Reorder()
			end
		end)
		
		optSubCol:OnUpdate(function(col)
			if type(col)~="table" or type(col.Get)~="function" then
				MBM:CheckThisColorOpt(optSubCol)
			else
				color = col
			end
		end)
		
		optSubWid:OnUpdate(function(wid)
			if type(wid)~="number" or wid<1 then
				optSubWid:Set(4)
			else
				width = wid
			end
		end)
		
		function MBM:OnUnitCreated(unit)
			if unit and unit:GetType() == "Pickup" then
				local playerName = GameLib.GetPlayerUnit():GetName()
				if string.match(unit:GetName(), L["SubduePattern"]) == playerName then
					subUnit = unit
				end
			end
		end
		
		function MBM:OnUnitDestroyed(unit)
			if unit and subUnit and unit:GetId() == subUnit:GetId() then
				subUnit = nil
			end
		end
		
		function MBM:DrawSubdue()
			if subUnit then
				if subUnit:IsValid() then
					local cut, tar = self:UnitToVecs(subUnit)
					if cut then
						local col = color:Get(subUnit)
						self:Draw(cut, tar, col, width)
					end
				else
					subUnit = nil
				end
			end
		end
	end
	
	do --player circle
		local optPlayer = MRF:GetOption(ModOpt, "playerCircle")
		local optPlayerAct = MRF:GetOption(optPlayer, "activate")
		optPlayerAct:OnUpdate(function(act)
			if act == nil then
				optPlayerAct:Set(true)
			elseif act then
				executeThis.DrawPlayer = true
				MBM:Reorder()
			else
				executeThis.DrawPlayer = nil
				MBM:Reorder()
			end
		end)
		
		local color, optPlayerColor = nil, MRF:GetOption(optPlayer, "color"); MBM:RegisterColorOption(optPlayerColor)
		optPlayerColor:OnUpdate(function(col)
			if type(col)~="table" or type(col.Get)~="function" then
				MBM:CheckThisColorOpt(optPlayerColor)
			else
				color = col
			end
		end)
		
		local radius, optPlayerRadius = nil, MRF:GetOption(optPlayer, "radius")
		optPlayerRadius:OnUpdate(function(rad)
			if rad==nil or type(rad)~="number" or rad<1 then
				optPlayerRadius:Set(7)
			else
				radius = rad
			end
		end)
		
		local show0hp, opt0hpFilter = nil, MRF:GetOption(ModOpt, "filters", "activate0HealthFilter")
		opt0hpFilter:OnUpdate(function(val)
			show0hp = not val
		end)
		
		function MBM:DrawPlayer()
			local player = units[1] or GameLib.GetPlayerUnit()
			
			if player then
				local pHealth = self:GetHealthPercentage(player)
				if show0hp or pHealth>0 then
					col = color:Get(player, pHealth)
					frame:AddPixie( {
						strSprite = "WhiteCircle", cr = col,
						loc = { fPoints = { 0.5, 0.5, 0.5, 0.5 }, nOffsets = { -radius, -radius, radius, radius } }
					} )
				end
			end
		end
	end
	
	do --Normal and Foci	
		local optNormal = MRF:GetOption(ModOpt, "standardLines")
		local optHighlight = MRF:GetOption(ModOpt, "highlightLines")
		local optCutoff = MRF:GetOption(ModOpt, "cutoff")
		local optNormAct = MRF:GetOption(optNormal, "activate")
		local optHighAct = MRF:GetOption(optHighlight, "activate")
		local optCutAct = MRF:GetOption(optCutoff, "activate")
		local color, optNormColor = nil, MRF:GetOption(optNormal, "color");			MBM:RegisterColorOption(optNormColor)
		local hColor, optHighColor = nil, MRF:GetOption(optHighlight, "color");		MBM:RegisterColorOption(optHighColor)
		local width, optNormWidth = nil, MRF:GetOption(optNormal, "width")
		local extWidth, optExtWidth = nil, MRF:GetOption(optHighlight, "width")
		local hWidth, optHighWidth = nil, MRF:GetOption(optHighlight, "highlightWidth")
		local cutoffPerc, optCutPerc = nil, MRF:GetOption(optCutoff, "percentage")
		local cutoffMax, optCutMax = nil, MRF:GetOption(optCutoff, "maximum")
	
		optNormAct:OnUpdate(function(act) 
			if act == nil then
				optNormAct:Set(true)
			elseif act then
				executeThis.DrawNoNormal = nil
				executeThis.DrawNormal = true
				MBM:Reorder()
			else
				executeThis.DrawNoNormal = true
				executeThis.DrawNormal = nil
				MBM:Reorder()
			end
		end)
	
		optHighAct:OnUpdate(function(act)
			if act == nil then
				optHighAct:Set(true)
			elseif act then
				executeThis.DrawNoFoci = nil
				executeThis.DrawFoci = true
				MBM:Reorder()
			else
				executeThis.DrawNoFoci = true
				executeThis.DrawFoci = nil
				MBM:Reorder()
			end
		end)
	
		optCutAct:OnUpdate(function(act)
			if act==nil then
				optCutAct:Set(true)
			elseif act then
				function MBM:GetCutoff(vec2)
					local new = vec2*cutoffPerc
					if new:Length() > cutoffMax then
						new = new:NormalFast()*cutoffMax
					end
					return new
				end
			else
				local vec = Vector2.New(0,0)
				function MBM:GetCutoff()
					return vec
				end
			end
		end)
	
		optNormColor:OnUpdate(function(col)
			if type(col)~="table" or type(col.Get)~="function" then
				MBM:CheckThisColorOpt(optNormColor)
			else
				color = col
			end
		end)
		
		optHighColor:OnUpdate(function(col)
			if type(col)~="table" or type(col.Get)~="function" then
				MBM:CheckThisColorOpt(optHighColor)
			else
				hColor = col
			end
		end)
		
		optNormWidth:OnUpdate(function(wid)
			if type(wid)~="number" or wid<1 then
				optNormWidth:Set(4)
			else
				width = wid
			end
		end)
		
		optExtWidth:OnUpdate(function(wid)
			if type(wid)~="number" or wid<1 then
				optExtWidth:Set(6)
			else
				extWidth = wid
			end
		end)
		
		optHighWidth:OnUpdate(function(wid)
			if type(wid)~="number" or wid<1 then
				optHighWidth:Set(3)
			else
				hWidth = wid
			end
		end)
		
		optCutPerc:OnUpdate(function(cut)
			if type(cut)~="number" or cut<0 or cut>99 then
				optCutPerc:Set(20)
			else
				cutoffPerc = cut/100
			end
		end)
		
		optCutMax:OnUpdate(function(cut)
			if type(cut)~="number" or cut<0 then
				optCutMax:Set(150)
			else
				cutoffMax = cut
			end
		end)
		
		local tFoci = {} --[unit] = true/false (dependant by if he should draw normal or not)
		
		function MBM:DrawNormal()
			tFoci = {}
			for i=2,#units,1 do --i=1 is the player, so not included.
				if self:Filter(units[i]) then
					if self:IsFocus(units[i]) then
						tFoci[units[i]] = true
					else
						local cut, tar = self:UnitToVecs(units[i])
						if cut then
							local col = color:Get(units[i], self:GetHealthPercentage(units[i]))
							self:Draw(cut, tar, col, width)
						end
					end
				end
			end
		end
		
		function MBM:DrawNoNormal()
			tFoci = {}
			for i=2,#units,1 do --i=1 is the player, so not included.
				if self:Filter(units[i]) and self:IsFocus(units[i]) then
					tFoci[units[i]] = false
				end
			end
		end
		
		function MBM:DrawFoci()
			for unit in pairs(tFoci) do
				local cut, tar = self:UnitToVecs(unit)
				if cut then
					local prec = self:GetHealthPercentage(unit)
					self:Draw(cut, tar, color:Get(unit, prec), extWidth)
					self:Draw(cut, tar, hColor:Get(unit, prec), hWidth)
				end
			end
		end
		
		function MBM:DrawNoFoci()
			for unit, draw in pairs(tFoci) do
				if draw then
					local cut, tar = self:UnitToVecs(unit)
					if cut then
						local col = color:Get(unit, self:GetHealthPercentage(unit))
						self:Draw(cut, tar, col, width)
					end
				end
			end
		end
	end
	
	do --Filtering
		local optFiltering = MRF:GetOption(ModOpt, "filters")
		local optAct0hp = MRF:GetOption(optFiltering, "activate0HealthFilter")
		local optActDist = MRF:GetOption(optFiltering, "activateDistanceFilter")
		local rangeSq, optLenDist = nil, MRF:GetOption(optFiltering, "lengthDistanceFilter")
		
		local pPos = nil --stored as Vector3 for easy Maths!
		local filters = {} -- [func] = true
		
		local function filter0hp(unit) -- returns true, if 'should be filtered out'
			local hp = unit:GetHealth()
			if hp==0 or not hp then return true end
			return false
		end
		
		local function filterRange(unit) -- returns true, if 'should be filtered out'
			if not pPos then return true end --if we do not have a position, we assume we are not in range.
			local pos = unit:GetPosition()
			if not pos then return true end --we do not get a position? - not in range!
			return (Vector3.New(pos)-pPos):LengthSq() > rangeSq --priorize to not root the length. because Math is hard! :)
		end
		
		optAct0hp:OnUpdate(function(act)
			if act==nil then
				optAct0hp:Set(true)
			elseif act then
				filters[filter0hp] = true
			else
				filters[filter0hp] = nil
			end
		end)
		
		optActDist:OnUpdate(function(act)
			if act==nil then
				optActDist:Set(true)
			elseif act then
				filters[filterRange] = true
				executeThis.UpdatePlayerPos = true
				MBM:Reorder()
			else
				filters[filterRange] = nil
				executeThis.UpdatePlayerPos = nil
				MBM:Reorder()
			end
		end)
		
		optLenDist:OnUpdate(function(len)
			if type(len)~="number" or len<0 then
				optLenDist:Set(70)
			else
				rangeSq=len^2
			end
		end)
		
		function MBM:UpdatePlayerPos()
			local player = GameLib.GetPlayerUnit()
			if not player then return end
			local pos = player:GetPosition()
			if not pos then return end
			pPos = Vector3.New(pos)
		end
		
		function MBM:Filter(unit) --return true, if a unit passed all filters.
			for func in pairs(filters) do
				if func(unit) then
					return false
				end
			end
			return true
		end
	end
	
	do --IS unit a Focus?
		local mbFocus = {}
		local altTar = nil --name of alternate target
		function MBM:IsFocus(unit)
			local name = unit:GetName() or false
			if altTar and name==altTar then
				return true
			elseif mbFocus[name] then
				return true
			end
			return false
		end
		
		Apollo.RegisterSlashCommand("mbfocus", "MBFocus", self)
		function MBM:MBFocus()
			local tar = GameLib.GetTargetUnit()
			if not tar then return end
			local name = tar:GetName()
			if name then
				mbFocus[name] = not mbFocus[name]
			end
		end
		
		
		MBM.checkTimer = ApolloTimer.Create(1, true, "CheckAlternateTarget", self)
		function MBM:CheckAlternateTarget()
			local player = GameLib.GetPlayerUnit()
			if not player then return end
			local alt = player:GetAlternateTarget() 
			altTar = alt and alt:GetName() or nil
		end
		
	end
	
	function MBM:GetHealthPercentage(unit)
		local max = unit:GetMaxHealth()
		local cur = unit:GetHealth()
		if not cur then cur = 0 end
		if not max or max<1 then max = 1 end
		
		return cur/max
	end
	
	ModOpt:ForceUpdate() --at this point the saves of mrf are already loaded. We will not get an automatic Update from MRF's SaveHandler
	
	local L = MRF:Localize(
		{ --English
			["ttPlayerCircle"] = "Draws a Indicator in the very center of your Screen, showing your own Health.",
			["ttNormalLine"] = "Draws Lines from the Screen-Center towards every groupmember.",
			["ttHighlightLine"] = "Draws special Lines towards every highlighted groupmember(in focus or /mbfocus).",
			["ttHighlightColor"] = "The Color of the highlighted Part.",
			["ttExtendedWidth"] = "The width of the normally colored part.",
			["ttHighlightWidth"] = "The width of the highlight.",
			["ttCutoff"] = "This tries to free up some space in the screens center, by clipping off a percentage of each Line.",
			["ttDisarm"] = "Draws a Line towards your weapon, if disarmed.",
		},
		{ --German
			["ttPlayerCircle"] = "Zeichnet einen Kreis in die Mitte de Bildschirms, welcher das eigene Leben darstellt.",
			["ttNormalLine"] = "Zeichnet Linien vom Bildschirm-Mittelpunkt zu jedem Gruppenmitglied.",
			["ttHighlightLine"] = "Zeichnet spezielle Linien zu jedem hervorgehobenen Gruppenmitglied.(im Fokus oder /mbfocus)",
			["ttHighlightColor"] = "Die Farbe des hervorgehobenen Anteils.",
			["ttExtendedWidth"] = "Die Breite des normal gefärbten Anteils im hervorgehobenen Modus.",
			["ttHighlightWidth"] = "Die Breite der Hervorhebung.",
			["ttCutoff"] = "Dies versucht die Mitte des Bildschirms etwas frei zu halten, indem ein Teil von jeder Linie abgeschnitten wird.",
			["ttDisarm"] = "Zeichnet eine Linie zu deiner Waffe, falls du Entwaffnet wurdest.",
		
		
			["General"] = "Allgemeines",
			["Activate Module"] = "Modul Aktivieren",
			["Don't draw dead members"] = "Tote Spieler nicht zeichnen",
			["Don't draw members out of distance:"] = "Distanzierte Spieler nicht zeichnen:",
			["Players Circle"] = "Spieler Kreis",
			["Activate"] = "Aktivieren",
			["Color:"] = "Farbe:",
			["Radius:"] = "Radius:",
			["Normal Lines"] = "Normale Linien",
			["Thickness:"] = "Breite:",
			["Highlighted Lines"] = "Hervorgehobene Linien",
			["Highlight Thickness:"] = "Breite (Hervorhebung):",
			["Central Clipping"] = "Mittlere Beschränkung",
			["Percentage Clipping:"] = "Prozentualer Abschnitt:",
			["Maximal Clipping:"] = "Maximaler Abschnitt:",
			["Disarm"] = "Entwaffnung",
		},
		{ --French
		}
	)
	
	MRF:AddMainTab("Buddy")
	MRF:AddChildTab(L["General"], "Buddy", MBM, "GeneralSettings")
	MRF:AddChildTab(L["Players Circle"], "Buddy", MBM, "PlayerCircleSettings")
	MRF:AddChildTab(L["Normal Lines"], "Buddy", MBM, "NormalLinesSettings")
	MRF:AddChildTab(L["Highlighted Lines"], "Buddy", MBM, "HighlightLinesSettings")
	MRF:AddChildTab(L["Central Clipping"], "Buddy", MBM, "CentralClppingSettings")
	MRF:AddChildTab(L["Disarm"], "Buddy", MBM, "SubdueSettings")
	
	local function sizeParent(wnd)
		wnd:SetSprite("BK3:UI_BK3_Holo_InsetSimple")
		wnd:ArrangeChildrenVert()
		local children = wnd:GetChildren()
		local l,t,r,b = wnd:GetAnchorOffsets()
		wnd:SetAnchorOffsets(l,t,r,select(4,children[#children]:GetAnchorOffsets())+t)
		wnd:GetParent():RecalculateContentExtents()
	end
	
	function MBM:GeneralSettings(parent, name)
		local parent = MRF:LoadForm("SimpleTab", parent)
		parent:FindChild("Title"):SetText("Buddy - "..name)
		parent = parent:FindChild("Space")
		
		local actRow = MRF:LoadForm("HalvedRow", parent)
		MRF:applyCheckbox(actRow:FindChild("Left"), MRF:GetOption(ModOpt, "activated"), L["Activate Module"])
		local fil0hpRow = MRF:LoadForm("HalvedRow", parent)
		MRF:applyCheckbox(fil0hpRow:FindChild("Left"), MRF:GetOption(ModOpt, "filters", "activate0HealthFilter"), L["Don't draw dead members"])
		local filDistRow = MRF:LoadForm("HalvedRow", parent)
		MRF:applyCheckbox(filDistRow:FindChild("Left"), MRF:GetOption(ModOpt, "filters", "activateDistanceFilter"), L["Don't draw members out of distance:"])
		MRF:applySlider(filDistRow:FindChild("Right"), MRF:GetOption(ModOpt, "filters", "lengthDistanceFilter"), 0, 100, 1, false, false, true)
		
		sizeParent(parent)
		ModOpt:ForceUpdate()
	end
	
	function MBM:PlayerCircleSettings(parent, name)
		local parent = MRF:LoadForm("SimpleTab", parent)
		parent:FindChild("Title"):SetText("Buddy - "..name)
		parent = parent:FindChild("Space")
	
		local colors, colorTrans = MRF:GetBarColors()
	
		local actPlayerRow = MRF:LoadForm("HalvedRow", parent)
		MRF:LoadForm("QuestionMark", actPlayerRow:FindChild("Right")):SetTooltip(L["ttPlayerCircle"])
		MRF:applyCheckbox(actPlayerRow:FindChild("Left"), MRF:GetOption(ModOpt, "playerCircle", "activate"), L["Activate"])
		local colPlayerRow = MRF:LoadForm("HalvedRow", parent)
		colPlayerRow:FindChild("Left"):SetText(L["Color:"])
		MRF:applyDropdown(colPlayerRow:FindChild("Right"), colors, MRF:GetOption(ModOpt, "playerCircle", "color"), colorTrans)
		local radPlayerRow = MRF:LoadForm("HalvedRow", parent)
		radPlayerRow:FindChild("Left"):SetText(L["Radius:"])
		MRF:applySlider(radPlayerRow:FindChild("Right"), MRF:GetOption(ModOpt, "playerCircle", "radius"), 1, 20, 1, false, false, true)
		
	
		sizeParent(parent)
		ModOpt:ForceUpdate()
	end
	
	function MBM:NormalLinesSettings(parent, name)
		local parent = MRF:LoadForm("SimpleTab", parent)
		parent:FindChild("Title"):SetText("Buddy - "..name)
		parent = parent:FindChild("Space")
	
		local colors, colorTrans = MRF:GetBarColors()
		
		local actNormRow = MRF:LoadForm("HalvedRow", parent)
		MRF:LoadForm("QuestionMark", actNormRow:FindChild("Right")):SetTooltip(L["ttNormalLine"])
		MRF:applyCheckbox(actNormRow:FindChild("Left"), MRF:GetOption(ModOpt, "standardLines", "activate"), L["Activate"])
		local colNormRow = MRF:LoadForm("HalvedRow", parent)
		colNormRow:FindChild("Left"):SetText(L["Color:"])
		MRF:applyDropdown(colNormRow:FindChild("Right"), colors, MRF:GetOption(ModOpt, "standardLines", "color"), colorTrans)
		local widNormRow = MRF:LoadForm("HalvedRow", parent)
		widNormRow:FindChild("Left"):SetText(L["Thickness:"])
		MRF:applySlider(widNormRow:FindChild("Right"), MRF:GetOption(ModOpt, "standardLines", "width"), 1, 20, 1, false, false, true)
	
		sizeParent(parent)
		ModOpt:ForceUpdate()
	end
	
	function MBM:HighlightLinesSettings(parent, name)
		local parent = MRF:LoadForm("SimpleTab", parent)
		parent:FindChild("Title"):SetText("Buddy - "..name)
		parent = parent:FindChild("Space")
	
		local colors, colorTrans = MRF:GetBarColors()

		local actHighRow = MRF:LoadForm("HalvedRow", parent)
		MRF:LoadForm("QuestionMark", actHighRow:FindChild("Right")):SetTooltip(L["ttHighlightLine"])
		MRF:applyCheckbox(actHighRow:FindChild("Left"), MRF:GetOption(ModOpt, "highlightLines", "activate"), L["Activate"])
		local colHighRow = MRF:LoadForm("HalvedRow", parent)
		MRF:LoadForm("QuestionMark", colHighRow):SetTooltip(L["ttHighlightColor"])
		colHighRow:FindChild("Left"):SetText(L["Color:"])
		MRF:applyDropdown(colHighRow:FindChild("Right"), colors, MRF:GetOption(ModOpt, "highlightLines", "color"), colorTrans)
		local widExtRow = MRF:LoadForm("HalvedRow", parent)
		MRF:LoadForm("QuestionMark", widExtRow):SetTooltip(L["ttExtendedWidth"])
		widExtRow:FindChild("Left"):SetText(L["Thickness:"])
		MRF:applySlider(widExtRow:FindChild("Right"), MRF:GetOption(ModOpt, "highlightLines", "width"), 1, 20, 1, false, false, true)
		local widHighRow = MRF:LoadForm("HalvedRow", parent)
		MRF:LoadForm("QuestionMark", widHighRow):SetTooltip(L["ttHighlightWidth"])
		widHighRow:FindChild("Left"):SetText(L["Highlight Thickness:"])
		MRF:applySlider(widHighRow:FindChild("Right"), MRF:GetOption(ModOpt, "highlightLines", "highlightWidth"), 1, 20, 1, false, false, true)
	
		sizeParent(parent)
		ModOpt:ForceUpdate()
	end
	
	function MBM:CentralClppingSettings(parent, name)
		local parent = MRF:LoadForm("SimpleTab", parent)
		parent:FindChild("Title"):SetText("Buddy - "..name)
		parent = parent:FindChild("Space")

		local actCutRow = MRF:LoadForm("HalvedRow", parent)
		MRF:LoadForm("QuestionMark", actCutRow:FindChild("Right")):SetTooltip(L["ttCutoff"])
		MRF:applyCheckbox(actCutRow:FindChild("Left"), MRF:GetOption(ModOpt, "cutoff", "activate"), L["Activate"])
		local perCutRow = MRF:LoadForm("HalvedRow", parent)
		perCutRow:FindChild("Left"):SetText(L["Percentage Clipping:"])
		MRF:applySlider(perCutRow:FindChild("Right"), MRF:GetOption(ModOpt, "cutoff", "percentage"), 1, 99, 1, false, false, false)
		local maxCutRow = MRF:LoadForm("HalvedRow", parent)
		maxCutRow:FindChild("Left"):SetText(L["Maximal Clipping:"])
		MRF:applySlider(maxCutRow:FindChild("Right"), MRF:GetOption(ModOpt, "cutoff", "maximum"), 1, 300, 1, false, false, true)
	
		sizeParent(parent)
		ModOpt:ForceUpdate()
	end
	
	function MBM:SubdueSettings(parent, name)
		local parent = MRF:LoadForm("SimpleTab", parent)
		parent:FindChild("Title"):SetText("Buddy - "..name)
		parent = parent:FindChild("Space")

		local colors, colorTrans = MRF:GetNoBarColors()
		
		local actSubRow = MRF:LoadForm("HalvedRow", parent)
		MRF:LoadForm("QuestionMark", actSubRow:FindChild("Right")):SetTooltip(L["ttDisarm"])
		MRF:applyCheckbox(actSubRow:FindChild("Left"), MRF:GetOption(ModOpt, "subdue", "activate"), L["Activate"])
		local colSubRow = MRF:LoadForm("HalvedRow", parent)
		colSubRow:FindChild("Left"):SetText(L["Color:"])
		MRF:applyDropdown(colSubRow:FindChild("Right"), colors, MRF:GetOption(ModOpt, "subdue", "color"), colorTrans)
		local widSubRow = MRF:LoadForm("HalvedRow", parent)
		widSubRow:FindChild("Left"):SetText(L["Thickness:"])
		MRF:applySlider(widSubRow:FindChild("Right"), MRF:GetOption(ModOpt, "subdue", "width"), 1, 20, 1, false, false, true)
	
		sizeParent(parent)
		ModOpt:ForceUpdate()
	end
end
-----------------------------------------------------------------------------------------------
-- MBM Instance
-----------------------------------------------------------------------------------------------
local MBMInst = MBM:new()
MBMInst:Init()
